import { DependencyList, Dispatch, MutableRefObject } from "react";
import { DebounceOptions } from "perfect-debounce";
export const depsChanged: (a: DependencyList, b: DependencyList) => boolean;
export const shallowMerge: <T extends UnknownRecord, U extends T | Partial<T> | undefined>(init: T, ...objs: U[]) => T;
export const isFn: (x: unknown) => x is (...args: unknown[]) => unknown;
export const isPromise: <T>(x: T | Promise<T>) => x is Promise<T>;
export type Res<T, E = Error> = {
    ok?: T;
    err?: E;
};
export type ResLoading<T, E = Error> = Res<T, E> & {
    loading: boolean;
};
export type UnknownRecord = Record<string | number | symbol, unknown>;
export type UnknownFn = (...args: any) => any;
export type UnknownAsyncFn = (...args: any) => Promise<any>;
export type UnknownVoidFn = (...args: any) => void;
export type UnknownAsyncVoidFn = (...args: any) => Promise<void>;
export type EffectFn = () => void;
export type AsyncEffectFn = () => Promise<void>;
export type InferredFn<F extends UnknownFn> = (...args: Parameters<F>) => ReturnType<F>;
export type InferredAsyncFn<F extends UnknownAsyncFn> = (...args: Parameters<F>) => AsPromise<ReturnType<F>> & Promise<any>;
export type UnPromise<P> = P extends Promise<infer T> ? T : P;
export type AsPromise<P> = P extends Promise<infer T> ? Promise<T> : Promise<P>;
export type AsAsyncFn<F extends UnknownFn> = (...args: Parameters<F>) => AsPromise<ReturnType<F>>;
export type AsSyncFn<F extends UnknownFn> = (...args: Parameters<F>) => UnPromise<ReturnType<F>>;
export type AsVoidFn<F extends UnknownFn> = (...args: Parameters<F>) => void;
export type AsAsyncVoidFn<F extends UnknownFn> = (...args: Parameters<F>) => Promise<void>;
/**
 * `useAsyncCb` provides a sync callback for an async function
 * along with helpers to track the state.
 */
export const useAsyncCb: <F extends InferredAsyncFn<F>>(fn: F, deps?: DependencyList, onError?: (err: Error) => void) => [ResLoading<UnPromise<ReturnType<F>>>, AsVoidFn<F>, EffectFn];
/**
 * `useAsyncEffect` handles async `useEffect` methods with optional error handling.
 */
export const useAsyncEffect: <F extends AsyncEffectFn & InferredAsyncFn<F>>(fn: F, deps?: DependencyList, onError?: (err: Error) => void) => void;
/**
 * `useAutoReducer` gives a preconfigured `useReducer` hook which is enough
 * for most needs. The `state` is initialized to the given value, and the
 * `dispatch` function accepts fragments of `state` which are merged in.
 */
export const useAutoReducer: <S extends UnknownRecord>(initState: S) => [S, Dispatch<Partial<S> | ((prev: S) => Partial<S>)>];
/**
 * `useCurrentRef` escapes closures. `ref.current` will always be
 * the latest `val` even if `ref` is captured.
 */
export const useCurrentRef: <T>(val: T) => MutableRefObject<T>;
/**
 * `useDebounceVal` gives a value that delays updates to `val` by the duration `dur`.
 */
export const useDebounceVal: <T>(val: T, dur?: number) => T;
/**
 * `useDebounceCb` consolidates multiple calls tp `fn` within duration `dur` into one.
 * Calls the given function immediately, use `opts` to change behavior.
 */
export const useDebounceCb: <F extends InferredFn<F>>(fn: F, deps?: DependencyList, dur?: number, opts?: DebounceOptions) => (...args: Parameters<F>) => Promise<ReturnType<F>>;
/**
 * `useMutexCb` returns a callback that allows one call to `fn` and
 * ignores all subsequent calls until it completes.
 */
export const useMutexCb: <F extends InferredAsyncFn<F>>(fn: F, deps?: DependencyList) => AsAsyncVoidFn<F>;
/**
 * `useStaticMutexCb` is the same as `useMutexCb`, but the returned ref is wrapped
 * with `useStaticCb` instead of the standard `useCallback`.
 */
export const useStaticMutexCb: <F extends InferredAsyncFn<F>>(fn: F, deps?: DependencyList) => AsAsyncVoidFn<F>;
/**
 * `useStaticMemo` will _never_ recompute the value unless `deps` change,
 * even if React thinks it'd be funny.
 */
export const useStaticMemo: <T>(gen: () => T, deps?: DependencyList) => T;
/**
 * `useStaticCb` gives a ref to `fn` which always calls the latest `fn`,
 * but won't trigger rerenders if `fn` changes.
 */
export const useStaticCb: <F extends InferredFn<F>>(fn: F, deps?: DependencyList) => F;
/**
 * `useStaticEffect` will be called once and only once,
 * without complaints about incomplete deps arrays.
 */
export const useStaticEffect: (fn: () => void, deps?: DependencyList) => void;
/**
 * `useLocalStore` provides persistent state from local browser storage.
 */
export const useLocalStore: <T>(key: string, initVal?: T) => [ResLoading<T>, (newVal?: T) => Promise<void>];

//# sourceMappingURL=index.d.ts.map
